{{
----------------------------------------------------------------------------------------
File: FullDuplexSerialPlus.spin
Version: 1.3
Copyright (c) 2009 Parallax, Inc.
See end of file for terms of use.

Original author: Chip Gracey
Original object: FullDuplexSerial.spin v1.1
----------------------------------------------------------------------------------------
}}

{
Modifications/additions by Andy Lindsay: 
v1.1 
 - Documentation comments
 - Parallax Serial Terminal constants
 - GetStr method
 - WaitStr method
 - Methods converting text strings into numeric values:
     o GetDec
     o StrToDec
     o GetHex
     o StrToHex
     o GetBin
     o StrToBin
v1.2     
 - Methods for Exchanging Values with PC Software     
     o GetLong
     o PutLong
v1.21 Cursor positioning methods added
 - CursorHome
 - CursorXY(x, y)
 - CursorLeft
 - CursorRight
 - CursorDown
 - RingBell
 - Backspace
 - TabRight
 - LineFeed
 - ClearEndOfLine
 - ClearDown
 - CarriageReturn
 - CursorX(x)
 - CursorY(y)
 - ClearScreen
 v1.22
 - StartPST
 v1.3
 - Fix buffer size bug, thanks Chuck!
}
  
CON                                          ''
''Parallax Serial Terminal Control Character Constants
''----------------------------------------------------
  HOME     =   1                             ''HOME     =   1          
  CRSRXY   =   2                             ''CRSRXY   =   2          
  CRSRL    =   3                             ''CRSRL    =   3          
  CRSRR    =   4                             ''CRSRR    =   4          
  CRSRUP   =   5                             ''CRSRUP   =   5          
  CRSRDN   =   6                             ''CRSRDN   =   6          
  BELL     =   7                             ''BELL     =   7          
  BKSP     =   8                             ''BKSP     =   8          
  TAB      =   9                             ''TAB      =   9          
  LF       =   10                            ''LF       =   10         
  CLREOL   =   11                            ''CLREOL   =   11         
  CLRDN    =   12                            ''CLRDN    =   12         
  CR       =   13                            ''CR       =   13         
  CRSRX    =   14                            ''CRSRX    =   14         
  CRSRY    =   15                            ''CRSRY    =   15         
  CLS      =   16                            ''CLS      =   16          

CON

   BUFFER_LENGTH = 128          'can be 2, 4, 8, 16, 32, 64, 128, 256
   BUFFER_MASK   = BUFFER_LENGTH - 1

VAR

  long  cog                     'cog flag/id

  long  rx_head                 '9 contiguous longs
  long  rx_tail
  long  tx_head
  long  tx_tail
  long  rx_pin
  long  tx_pin
  long  rxtx_mode
  long  bit_ticks
  long  buffer_ptr
                     
  byte  rx_buffer[BUFFER_LENGTH]'transmit and receive buffers
  byte  tx_buffer[BUFFER_LENGTH]  


PUB StartPST(baudrate) : okay

'' Starts communicating with the Parallax Serial Terminal using the Propeller's
'' programming connection.  Waits 1 second for connection, then clears screen.
''   baudrate - bits per second.  Make sure it matches the Parallax Serial
''              Terminal's Baud Rate field.
''   okay     - returns nonzero, or false if no cog is available

  okay := start(31, 30, 0, baudrate)
  waitcnt(clkfreq + cnt)                        ' Wait 1 s for PST
  ClearScreen                                   ' Clear display

PUB Start(rxpin, txpin, mode, baudrate) : okay

{{ Starts serial driver in a new cog
   rxpin - input receives signals from peripheral's TX pin
   txpin - output sends signals to peripheral's RX pin
   mode  - bits in this variable configure signaling
              bit 0 inverts rx
              bit 1 inverts tx
              bit 2 open drain/source tx
              bit 3 ignor tx echo on rx
   baudrate - bits per second
   okay - returns nonzero false if no cog is available.}}

  stop
  longfill(@rx_head, 0, 4)
  longmove(@rx_pin, @rxpin, 3)
  bit_ticks := clkfreq / baudrate
  buffer_ptr := @rx_buffer
  okay := cog := cognew(@entry, @rx_head) + 1

PUB Stop

  '' Stops serial driver - frees a cog

  if cog
    cogstop(cog~ - 1)
  longfill(@rx_head, 0, 9)

PUB Tx(txbyte)

  '' Sends byte (may wait for room in buffer)

   repeat until (tx_tail <> ((tx_head + 1) & BUFFER_MASK))
   tx_buffer[tx_head] := txbyte
   tx_head := (tx_head + 1) & BUFFER_MASK

  if rxtx_mode & %1000
    rx

PUB Rx : rxbyte

  '' Receives byte (may wait for byte)
  '' rxbyte returns $00..$FF

  repeat while (rxbyte := rxcheck) < 0

PUB Str(stringptr)

  '' Send zero terminated string that starts at the stringptr memory address

  repeat strsize(stringptr)
    tx(byte[stringptr++])

PUB GetStr(stringptr) | index

    '' Gets a carriage return terminated string and stores it (starting at stringptr)
    '' as a zero terminated string
    
    index~
    repeat until ((byte[stringptr][index++] := rx) == 13)
    byte[stringptr][--index]~

PUB WaitStr(stringptr, timeout) : success | index, tstart

{{ Wait for a string.
   Parameters: 
     stringptr - points to the comparison string
     timeout   - number of clock ticks to wait
   Returns:
     success   - true if string found or false if timeout exceeded}}
        
  tstart := cnt
  index~
  repeat until byte[stringptr][index] == 0
    if (cnt - tstart) > timeout
      return
    if byte[stringptr][index] == rx
      index++
    else
      index~
  success~~

PUB Dec(value) | i

'' Prints a decimal number

  if value < 0
    -value
    tx("-")

  i := 1_000_000_000

  repeat 10
    if value => i
      tx(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      tx("0")
    i /= 10


PUB GetDec : value | tempstr[11]

    '' Gets decimal character representation of a number from the terminal
    '' Returns the corresponding value

    GetStr(@tempstr)
    value := StrToDec(@tempstr)    

PUB Bin(value, digits)

  '' Sends the character representation of a binary number to the terminal.

  value <<= 32 - digits
  repeat digits
    tx((value <-= 1) & 1 + "0")

PUB GetBin : value | tempstr[11]

  '' Gets binary character representation of a number from the terminal
  '' Returns the corresponding value
   
  GetStr(@tempstr)
  value := StrToBin(@tempstr)    
   
PUB Hex(value, digits)

  '' Print a hexadecimal number

  value <<= (8 - digits) << 2
  repeat digits
    tx(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB GetHex : value | tempstr[11]

    '' Gets hexadecimal character representation of a number from the terminal
    '' Returns the corresponding value

    GetStr(@tempstr)
    value := StrToHex(@tempstr)    

PUB GetLong : value | index

  '' Receives four serial bytes that comprise an int in little endian format.
  '' Returns an int.
  
  repeat index from 0 to 3
    byte[@value][index]:=rx

PUB PutLong(value) | index

  '' Transmits and int value parameter as four bytes in little endian format.
  
  repeat index from 0 to 3
    tx(byte[@value][index])

PUB CursorHome

  '' Send cursor to home position (top-left).
  
  tx(HOME)
  
PUB CursorXY(x, y)

  '' Place cursor x characters over and y characters down
  '' (from top-left).
  
  tx(CRSRXY)
  tx(x)
  tx(y)
  
PUB CursorLeft

  '' Move cursor left.
  
  tx(CRSRL)
  
PUB CursorRight

  '' Move cursor right.
  
  tx(CRSRR)
  
PUB CursorDown

  '' Move cursor down.
  
  tx(CRSRDN)
  
PUB RingBell

  '' Play bell tone on PC.
  
  tx(BELL)
  
PUB Backspace

  '' Backspace.
  
  tx(BKSP)
  
PUB TabRight

  '' Tab to the right.
  
  tx(TAB)
  
PUB LineFeed

  '' Line feed.
  
  tx(LF)
  
PUB ClearEndOfLine

  '' Clear to end of line.
  
  tx(CLREOL)
  
PUB ClearDown

  '' Clear all lines below cursor.
  
  tx(CLRDN)
  
PUB CarriageReturn

  '' Carriage return (includes line feed).
  
  tx(CR)
  
PUB CursorX(x)

  '' Move cursor x spaces from left margin.
  tx(CRSRX)
  tx(x)
  
PUB CursorY(y)

  '' Move cursor y spaces from top.
  tx(CRSRY)
  tx(y)
  
PUB ClearScreen

  '' Clear display and place cursor at top-left.
  
  tx(CLS)

PUB RxFlush

  '' Flush receive buffer

  repeat while rxcheck => 0
    
PUB RxCheck : rxbyte

  '' Check if byte received (never waits)
  '' rxbyte returns -1 if no byte received, $00..$FF if byte

  rxbyte--
  if rx_tail <> rx_head
    rxbyte := rx_buffer[rx_tail]
    rx_tail := (rx_tail + 1) & BUFFER_MASK

PUB RxTime(ms) : rxbyte | t

  '' Wait ms milliseconds for a byte to be received
  '' returns -1 if no byte received, $00..$FF if byte

  t := cnt
  repeat until (rxbyte := rxcheck) => 0 or (cnt - t) / (clkfreq / 1000) > ms

PUB StrToBin(stringptr) : value | char, index

  '' Converts a zero terminated string representaton of a binary number to a value
   
  value := index := 0
  repeat until ((char := byte[stringptr][index++]) == 0)
     if char => "0" and char =< "1"
        value := value * 2 + (char - "0")
  if byte[stringptr] == "-"
     value := - value
     
PUB StrToDec(stringptr) : value | char, index, multiply

    '' Converts a zero terminated string representation of a decimal number to a value

    value := index := 0
    repeat until ((char := byte[stringptr][index++]) == 0)
       if char => "0" and char =< "9"
          value := value * 10 + (char - "0")
    if byte[stringptr] == "-"
       value := - value
       
PUB StrToHex(stringptr) : value | char, index

    '' Converts a zero terminated string representaton of a hexadecimal number to a value

    value := index := 0
    repeat until ((char := byte[stringptr][index++]) == 0)
       if (char => "0" and char =< "9")
          value := value * 16 + (char - "0")
       elseif (char => "A" and char =< "F")
          value := value * 16 + (10 + char - "A")
       elseif(char => "a" and char =< "f")   
          value := value * 16 + (10 + char - "a")
    if byte[stringptr] == "-"
       value := - value
   
DAT

'***********************************
'* Assembly language serial driver *
'***********************************

                        org
'
'
' Entry
'
entry                   mov     t1,par                'get structure address
                        add     t1,#4 << 2            'skip past heads and tails

                        rdlong  t2,t1                 'get rx_pin
                        mov     rxmask,#1
                        shl     rxmask,t2

                        add     t1,#4                 'get tx_pin
                        rdlong  t2,t1
                        mov     txmask,#1
                        shl     txmask,t2

                        add     t1,#4                 'get rxtx_mode
                        rdlong  rxtxmode,t1

                        add     t1,#4                 'get bit_ticks
                        rdlong  bitticks,t1

                        add     t1,#4                 'get buffer_ptr
                        rdlong  rxbuff,t1
                        mov     txbuff,rxbuff
                        add     txbuff,#BUFFER_LENGTH

                        test    rxtxmode,#%100  wz    'init tx pin according to mode
                        test    rxtxmode,#%010  wc
        if_z_ne_c       or      outa,txmask
        if_z            or      dira,txmask

                        mov     txcode,#transmit      'initialize ping-pong multitasking
'
'
' Receive
'
receive                 jmpret  rxcode,txcode         'run chunk of tx code, then return

                        test    rxtxmode,#%001  wz    'wait for start bit on rx pin
                        test    rxmask,ina      wc
        if_z_eq_c       jmp     #receive

                        mov     rxbits,#9             'ready to receive byte
                        mov     rxcnt,bitticks
                        shr     rxcnt,#1
                        add     rxcnt,cnt                          

:bit                    add     rxcnt,bitticks        'ready next bit period

:wait                   jmpret  rxcode,txcode         'run chunk of tx code, then return

                        mov     t1,rxcnt              'check if bit receive period done
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #:wait

                        test    rxmask,ina      wc    'receive bit on rx pin
                        rcr     rxdata,#1
                        djnz    rxbits,#:bit

                        shr     rxdata,#32-9          'justify and trim received byte
                        and     rxdata,#$FF
                        test    rxtxmode,#%001  wz    'if rx inverted, invert byte
        if_nz           xor     rxdata,#$FF

                        rdlong  t2,par                'save received byte and inc head
                        add     t2,rxbuff
                        wrbyte  rxdata,t2
                        sub     t2,rxbuff
                        add     t2,#1
                        and     t2,#BUFFER_MASK
                        wrlong  t2,par

                        jmp     #receive              'byte done, receive next byte
'
'
' Transmit
'
transmit                jmpret  txcode,rxcode         'run chunk of rx code, then return

                        mov     t1,par                'check for head <> tail
                        add     t1,#2 << 2
                        rdlong  t2,t1
                        add     t1,#1 << 2
                        rdlong  t3,t1
                        cmp     t2,t3           wz
        if_z            jmp     #transmit

                        add     t3,txbuff             'get byte and inc tail
                        rdbyte  txdata,t3
                        sub     t3,txbuff
                        add     t3,#1
                        and     t3,#BUFFER_MASK
                        wrlong  t3,t1

                        or      txdata,#$100          'ready byte to transmit
                        shl     txdata,#2
                        or      txdata,#1
                        mov     txbits,#11
                        mov     txcnt,cnt

:bit                    test    rxtxmode,#%100  wz    'output bit on tx pin 
                        test    rxtxmode,#%010  wc    'according to mode
        if_z_and_c      xor     txdata,#1
                        shr     txdata,#1       wc
        if_z            muxc    outa,txmask        
        if_nz           muxnc   dira,txmask
                        add     txcnt,bitticks        'ready next cnt

:wait                   jmpret  txcode,rxcode         'run chunk of rx code, then return

                        mov     t1,txcnt              'check if bit transmit period done
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #:wait

                        djnz    txbits,#:bit          'another bit to transmit?

                        jmp     #transmit             'byte done, transmit next byte
'
'
' Uninitialized data
'
t1                      res     1
t2                      res     1
t3                      res     1

rxtxmode                res     1
bitticks                res     1

rxmask                  res     1
rxbuff                  res     1
rxdata                  res     1
rxbits                  res     1
rxcnt                   res     1
rxcode                  res     1

txmask                  res     1
txbuff                  res     1
txdata                  res     1
txbits                  res     1
txcnt                   res     1
txcode                  res     1

{{
+--------------------------------------------------------------------------------------+
                           TERMS OF USE: MIT License                                                                                              
+--------------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy of this  
software and associated documentation files (the "Software"), to deal in the Software  
without restriction, including without limitation the rights to use, copy, modify,    
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to    
permit persons to whom the Software is furnished to do so, subject to the following   
conditions:                                                                                                                       
                                                                                                                                     
The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.                                              
                                                                                                                                      
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,   
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A         
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT    
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION     
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE        
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                                
+--------------------------------------------------------------------------------------+
}}